/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2017 Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* 
 * Program to do parallel quicksort.
 * Ben Moszkowski
 */

define par_quicksort(L) = {
    if |L| <= 1
    then empty
    else exists pivot : {
	quick_partition(L,pivot);
	sort_parts(L,pivot)
    }
}.

define sort_parts(L,pivot) = {
    exists Ready1,Ready2 : {
	halt (Ready1 and Ready2)
	and quicksort_process(Ready1, L[0 to pivot])
	and quicksort_process(Ready2, L[pivot+1 to |L|])
        }
    and stable L[pivot]
}.

define quicksort_process(Ready,L) = process {
    {
	par_quicksort(L) and
	keep(Ready=false)
    };
    {   Ready= true and 
	stable Ready and stable L
    }
}.

define monitor(L) = {
    exists Tag_list : {
	fixed_list(Tag_list,|L|) and
	forall i<|L| : {
	    Tag_list[i] is (if i=L[i] then 1 else 0)
	} and
	box format("List=%t  Tag_list=%t\n",L,Tag_list)
    }
}.

/*  test of quicksort */
define sort_test(init_vector) = {
    exists L : {
	fixed_list(L,|init_vector|)
	and L=init_vector
	and par_quicksort(L)
        and monitor(L)
    }
}.

define sort_test1 = [2,0,1].

define sort_test2 = [2,4,9,1,0,10,6,8,7,5,3].

define sort_test3 = [13,7,5,3,1,9,8,6,2,0,12,11,4,10].

