/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *
 * Copyright (C) 2012-2017  Amin Mohammed El-kustaban
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

define p_free           = 0.
define p_busy           = 1.


define tr_ideal         = 4.
define t_pending        = 3.
define t_commit         = 2.
define t_abort          = 1.

define read             = 0.
define write            = 1.
define try_commit       = 2.
define try_abort        = 3.
define res_read_OL      = 4.
define res_read_OG      = 5.
define res_write_OK     = 6.
define res_commit       = 7.
define res_abort        = 8.

define nlocations      = 7.
define initial_value   = 0.

define nprocesses       = 3.
define ntransactions    = 2.
define nevents          = 15.

define no_event        = 1000.
define no_value        = -1.
define a               = 2.
define b               = 3.
define c               = 4.
define t0              = 0.
define e0              = 0.



load "../library/conversion".
load "../library/tcl".
/* tcl trans */
/*---------------------------------------------------------------------------------*/
/* run */ define initial_process() = {


 exists Process, Transaction, Event,GlobalMemory,
 LocalMemory,LocalMemoryS, Tr_begin_time,Tr_end_time,Event_object,Event_value,Timer : {


 list(Process, nprocesses) and stable(struct(Process)) and
 (forall i<nprocesses : Process[i] = p_free) and


 list(Transaction,nprocesses ) and stable(struct(Transaction)) and
 forall i<nprocesses: {
  list(Transaction[i], ntransactions) and stable(struct(Transaction[i]))
 } and
 (forall i<nprocesses :
  forall j<ntransactions :
	  Transaction[i][j]=tr_ideal
 ) and

 list(Tr_begin_time,nprocesses ) and stable(struct(Tr_begin_time)) and
 forall i<nprocesses: {
  list(Tr_begin_time[i], ntransactions) and stable(struct(Tr_begin_time[i]))
 } and
 (forall i<nprocesses :
	forall j<ntransactions :
	 Tr_begin_time[i][j]=1000

 ) and

 list(Tr_end_time,nprocesses ) and stable(struct(Tr_end_time)) and
 forall i<nprocesses: {
  list(Tr_end_time[i], ntransactions) and stable(struct(Tr_end_time[i]))
 } and
 (forall i<nprocesses :
	(forall j<ntransactions :
	  Tr_end_time[i][j] = 0              /* ----it does not consider initial value. I have to initialize it in a single state----*/

	)
 ) and

 list(Event,nprocesses ) and stable(struct(Event)) and
  forall i<nprocesses: {
	list(Event[i], ntransactions) and stable(struct(Event[i])) and
	  forall j<ntransactions: {
		list(Event[i][j], nevents) and stable(struct(Event[i][j]))
  }
	 } and

  (forall i<nprocesses :
	 (forall j<ntransactions :
		 (forall k<nevents :
			Event[i][j][k] = no_event
		 )
	)
 ) and

 list(Event_object,nprocesses ) and stable(struct(Event_object)) and
  forall i<nprocesses: {
	list(Event_object[i], ntransactions) and stable(struct(Event_object[i])) and
	  forall j<ntransactions: {
		list(Event_object[i][j], nevents) and stable(struct(Event_object[i][j]))
  }
	 } and

  (forall i<nprocesses :
	 (forall j<ntransactions :
		 (forall k<nevents :
			Event_object[i][j][k] = no_event
		 )
	)
 ) and

 list(Event_value,nprocesses ) and stable(struct(Event_value)) and
  forall i<nprocesses: {
	list(Event_value[i], ntransactions) and stable(struct(Event_value[i])) and
	  forall j<ntransactions: {
		list(Event_value[i][j], nevents) and stable(struct(Event_value[i][j]))
  }
	 } and

  (forall i<nprocesses :
	 (forall j<ntransactions :
		 (forall k<nevents :
			Event_value[i][j][k] = no_event
		 )
	)
 ) and


list(GlobalMemory, nlocations) and stable(struct(GlobalMemory)) and
 (forall i<nlocations : GlobalMemory[i] = initial_value) and

list(LocalMemory,nprocesses ) and stable(struct(LocalMemory)) and
 forall i<nprocesses: {
  list(LocalMemory[i], nlocations) and stable(struct(LocalMemory[i]))
 } and
 (forall i<nprocesses :
	(forall j<nlocations :
		 LocalMemory[i][j] = initial_value
	)
 ) and

 list(LocalMemoryS,nprocesses ) and stable(struct(LocalMemoryS)) and
 forall i<nprocesses: {
  list(LocalMemoryS[i], nlocations) and stable(struct(LocalMemoryS[i]))
 } and
 (forall i<nprocesses :
	(forall j<nlocations :
		 LocalMemoryS[i][j] = initial_value
	)
 ) and

 Timer = 0 and
 /* Timer = k : the time is currently k
 */

/*---------------------------------------------------------------------------------*/

 /* The Process states
  *
 */

/*---------------------------------------------------------------------------------*/

define process_idle(X) = {
	 skip and
	 stable(Process[X]) and
	 format("Process %t ....................remains in current state\n",X)

 } and

define process_busy(X) = {
	skip and
	 if Process[X]=p_free
		then { Process[X]:=p_busy
				/* format("Process %t .....will invoke a transaction\n",X)*/}
		 else
			 stable(Process[X])

 } and

define process_free(X) = {
	skip and
	 if Process[X]=p_busy
		then  {Process[X]:=p_free }
		 else
			 stable(Process[X])

 } and

  /*---------------------------------------------------------------------------------*/
 /*
  *
  * The Transactions states
  *
 */

 /*---------------------------------------------------Idle------------------------------*/
define transactionEvent_idle(X,Y) = {
  skip and
  tr_state_stable(X,Y)and
  (forall e<nevents: Event_idle(X,Y,e))
  /*format("Transaction %t at process %t .. remains in current state\n",Y,X)*/
 } and

 define tr_state_stable(X,Y) = {
  skip and
  stable(Transaction[X][Y]) and
  stable(Tr_begin_time[X][Y])and
  stable(Tr_end_time[X][Y])

 } and

 define Event_idle(X,Y,E) = {

  stable(Event[X][Y][E]) and
  stable(Event_object[X][Y][E]) and
  stable(Event_value[X][Y][E])

 } and

 define GlobalMemory_idle()={
 skip and
 (forall i<nlocations : stable GlobalMemory[i])
 }and

 define LocalMemory_idle(X)={
 skip and
 (forall i<nlocations : {stable LocalMemoryS[X][i] and
								 stable LocalMemory[X][i]})
 }and

 define itsParts_idle(X)={

	LocalMemory_idle(X)and
  (forall j<ntransactions: transactionEvent_idle(X,j))

 }and
 /*-----------------------------------------------Test----------------------------------*/

 define checkM(X)={
 (forall j<nlocations: if j~=1000 	then {format("GMemory[%t]=%t LMemory[%t][%t]=[%t]\n",j,GlobalMemory[j],X,j,LocalMemory[X][j])})

 }and


 /*-----------------------------------------&&&&&&&&&&&&&........The Main Functions.......&&&&&&&&&------------------------------------------------*/

									  /*-----------------&&&&&&&&.......Transaction Begin.......&&&&&&-----------------------------------*/


define transaction_begin(X,Y,E,Operation,Obj,Value)= {
			skip and

					  if  Process[X]=p_free and Transaction[X][Y]= tr_ideal
							then{ process_busy(X) and
									Transaction[X][Y]:= t_pending and
									Tr_begin_time[X][Y]:=Timer and
									Event[X][Y][E]:=Operation  and
									Event_object[X][Y][E]:=Obj and
									if Operation= write then Event_value[X][Y][E]:=Value
															  else{stable Event_value[X][Y][E]} and
									stable(Tr_end_time[X][Y])and
									LocalMemory_idle(X) and
									if check_concurent_TryC(X,Y,E,nprocesses-1)=0
										then { GlobalMemory_idle()}  and

									out_operation(X,Y,Operation,Obj,Value)
								 }
							else stable (Transaction[X][Y])


} and

							  /*--------------------&&&&&&&&.........Transaction Continue.......&&&&&&-----------------------------------*/


define transaction_continue(X,Y,E,Operation,Obj,Value)= {
		skip and
				  if  Transaction[X][Y]=t_pending
						then {Event[X][Y][E]:=Operation and
								Event_object[X][Y][E]:=Obj and
								if Operation= write then Event_value[X][Y][E]:=Value
														  else{stable Event_value[X][Y][E]} and
								 LocalMemory_idle(X) and
								 tr_state_stable(X,Y)  and
								 stable Process[X] and
								if check_concurent_TryC(X,Y,E,nprocesses-1)=0
									then { GlobalMemory_idle()} and

								 out_operation(X,Y,Operation,Obj,Value)
								}
						else  stable(Transaction[X][Y])
} and

						/*--------------------&&&&&&&&.........Transaction End.......&&&&&&-----------------------------------*/


define transaction_end(X,Y,E,Tr_end)= {
		 skip and
					if  Transaction[X][Y] = t_pending
						  then {Event[X][Y][E]:=Tr_end and
								  Event_value[X][Y][E]:=no_value and
								  Event_object[X][Y][E]:=no_value and
								  Tr_end_time[X][Y]:=Timer and
								  stable Tr_begin_time[X][Y] and
								  LocalMemory_idle(X) and
								  stable Transaction[X][Y] and
								  stable Process[X] and
								  if check_concurent_TryC(X,Y,E,nprocesses-1)=0
									  then { GlobalMemory_idle()}and
										out_operation(X,Y,Tr_end,0,0)
								 }
						  else  stable(Transaction[X][Y])
} and



						/*--------------------&&&&&&&&.........Transaction Response.......&&&&&&-----------------------------------*/


define transaction_response(X,Y,E)= {

 skip and


		  if  Transaction[X][Y]=t_pending
				then{   if   Event[X][Y][E-1]= try_commit

 /*-----Response of TryC operation-----*/
								then  {transaction_response_TryC(X,Y,E,nprocesses-1)}
								else{   if check_concurent_TryC(X,Y,E,nprocesses-1)=0
												then { GlobalMemory_idle()}and

										 if   Event[X][Y][E-1]= write
 /*----Response of write operation----*/
									 then{  Event[X][Y][E]:=res_write_OK and
												Event_object[X][Y][E]:=Event_object[X][Y][E-1]   and
												Event_value[X][Y][E]:=Event_value[X][Y][E-1] and
											  (forall i<nlocations : if i=Event_object[X][Y][E-1]
																		  then {LocalMemory[X][i]:= Event_value[X][Y][E-1] and
																				  LocalMemoryS[X][i]:=1}
																		  else {stable LocalMemory[X][i] and
																					stable LocalMemoryS[X][i]}
																					) and
											tr_state_stable(X,Y)  and
											stable Process[X] and
											out_operation(X,Y,res_write_OK,0,0)
										 }
									else if   Event[X][Y][E-1]= try_abort
/*---- Response of TryA operation-----*/
											  then {
														 Event[X][Y][E]:=res_abort and
														 transaction_abort(X,Y,E)
													 }
											  else if   Event[X][Y][E-1]= read
/* ----Response of read operation-----*/
																		 then transaction_response_Read(X,Y,E,E-1)
																		 else stable(Event[X][Y][E-1])


										 }
						}
						 else stable(Transaction[X][Y])
 } and




		/*--------------------&&&&&&&&.........Transaction Commit and Abort.......&&&&&&-----------------------------------*/



define transaction_commit(X,Y,E)={
		skip and
				  if  Transaction[X][Y] = t_pending
						then {Transaction[X][Y]:=t_commit and
								process_free(X) and

								(forall i<nlocations : if  LocalMemoryS[X][i]=1
																			 then GlobalMemory[i]:=LocalMemory[X][i]
																			 else stable GlobalMemory[i]) and

								(forall i<nlocations : {LocalMemory[X][i]:=0 and  LocalMemoryS[X][i]:=0})  and


								Event_object[X][Y][E]:=no_value and
								Event_value[X][Y][E]:= no_value and
								stable Tr_begin_time[X][Y] and
								stable Tr_end_time[X][Y] and
								out_operation(X,Y,res_commit,0,0)

								}
						else stable(Transaction[X][Y])
}and

 define transaction_abort(X,Y,E)={
		skip and
				  if  Transaction[X][Y] = t_pending
						then {Transaction[X][Y]:=t_abort and
								process_free(X) and
								(forall i<nlocations : {LocalMemory[X][i]:=0 and LocalMemoryS[X][i]:=0}
								 )  and

								Event_object[X][Y][E]:=no_value and
								Event_value[X][Y][E]:= no_value and
								stable Tr_begin_time[X][Y]  and
								stable Tr_end_time[X][Y] and
								if check_concurent_TryC(X,Y,E,nprocesses-1)=0
									 then { GlobalMemory_idle()} 
										else {format(" No GM stable \n")}and
								out_operation(X,Y,res_abort,0,0)
								}
						else stable(Transaction[X][Y])
}and

/*---------------------------------------------------------------------------------*/

					  /*--------------------&&&&&&&&.......Functions regarding..Transaction Response.....&&&&&&-----------------------------------*/



/*---------------------------------------------------------------------------------*/


 /*-------------------------------------------------------------------------------------------------*/
 define transaction_response_TryC(X,Y,E,n)={

	if n=-1 then {Event[X][Y][E]:=res_commit and
						transaction_commit(X,Y,E)

					 }
			 else	if (n~=X)
						then {    
								if (check_overlap_and_conflict(X,Y,n,ntransactions-1,E)=1)
								  then {Event[X][Y][E]:=res_abort and
										  transaction_abort(X,Y,E)
										}
								  else transaction_response_TryC(X,Y,E,n-1)
								}
						 else transaction_response_TryC(X,Y,E,n-1)

 } and


/*define  check_overlap_and_print(X,Y,i,j)={
	if j=-1 then format("NO over\n")
			  else    {  format(" beg[P%t T%t]=%t end[P%t T%t]=%t beg=%t end=%t\n",X,Y,Tr_begin_time[X][Y],i,j,Tr_end_time[i][j],Tr_begin_time[i][j],Tr_end_time[X][Y])and
						 if transaction_overlap(X,Y,i,j)  and  transaction_conflict_detect(X,Y,i,j,nevents-1)=1
						  then format(" overlap between [P%t T%t][P%t T%t]\n",X,Y,i,j)
						  else   check_overlap_and_print(X,Y,i,j-1)}


}and */



 define  check_overlap_and_conflict(X,Y,i,j,E)={
	if j=-1 then 0
			  else  if transaction_overlap(X,Y,i,j) and  transaction_conflict_detect(X,Y,i,j,nevents-1)=1 and contention_manager_abort(X,i,j,E)=1
						  then 1
						  else   check_overlap_and_conflict(X,Y,i,j-1,E)


}and

define  contention_manager_abort(X,i,j,E)={
	if  Transaction[i][j]=t_commit       /* whoever tries to commit first wins */
		then 1
			  else  if Transaction[i][j]=t_pending  and Event[i][j][E-1]=try_commit and X>i /* if two at the same time then priority*/
				 then 1
						  else   0


}and

/*---------------------------------------------------------------------------------*/

 define transaction_overlap(X,Y,i,j)={

 Tr_begin_time[X][Y]<=Tr_end_time[i][j] and Tr_begin_time[i][j]<=Tr_end_time[X][Y]


  }and

/*-------------------------------------------------------------------------------------------*/
 define transaction_conflict_detect(X,Y,i,j,e)={

	if e=-1 then 0
			  else  if (Event[X][Y][e]=read and  re_wr_conflict(X,Y,i,j,e,nevents-1)=1)or (Event[X][Y][e]=write and wr_reORwr_conflict(X,Y,i,j,e,nevents-1)=1)
						  then 1
						  else  transaction_conflict_detect(X,Y,i,j,e-1)
 }and
/*--------------1------*/

 define re_wr_conflict(X,Y,i,j,e,k)={
 if k=-1 then 0
			else if Event[i][j][k]=write and Event_object[X][Y][e]=Event_object[i][j][k]
				  then 1
				  else re_wr_conflict(X,Y,i,j,e,k-1)
 }and
/*--------------2------*/

 define wr_reORwr_conflict(X,Y,i,j,e,k)={
  if k=-1 then 0
			else if (Event[i][j][k]=write or Event[i][j][k]=read)and Event_object[X][Y][e]=Event_object[i][j][k]
				  then 1
				  else wr_reORwr_conflict(X,Y,i,j,e,k-1)

 }and
/*----------------*/

define check_concurent_TryC(X,Y,E,n)={

 if n=-1 then 0
			else if (n~=X and E~=0 and concurent_TryC(n,E,ntransactions-1)=1)
				  then 1
				  else check_concurent_TryC(X,Y,E,n-1)

}  and

/*----------------*/

define concurent_TryC(n,E,k)={

 if k=-1 then 0
			else if (Event[n][k][E-1]=try_commit)
				  then 1
				  else concurent_TryC(n,E,k-1)

}  and

/*---------------------------------------------------------------------------------*/
define transaction_response_Read(X,Y,E,n)={


					/*-----------------check if there is write for the object before the read in the same transaction---------*/

 if n=0 then { tr_state_stable(X,Y) and  stable Process[X]and LocalMemory_idle(X)and  /* if there is no abort*/
					Event[X][Y][E]:= res_read_OG  and
					Event_object[X][Y][E]:=Event_object[X][Y][E-1] and
					Event_value[X][Y][E]:=GlobalMemory[Event_object[X][Y][E-1]] and
					out_operation(X,Y,res_read_OG,Event_object[X][Y][E-1],GlobalMemory[Event_object[X][Y][E-1]])
				  }
			else  if      Event[X][Y][n]=write and Event_object[X][Y][n]=Event_object[X][Y][E-1]
								then{   tr_state_stable(X,Y) and stable Process[X]and LocalMemory_idle(X) and
										  Event[X][Y][E]:= res_read_OL and
										  Event_object[X][Y][E]:= Event_object[X][Y][E-1] and
										  Event_value[X][Y][E] := LocalMemory[X][Event_object[X][Y][ E-1]] and
											out_operation(X,Y,res_read_OL,Event_object[X][Y][E-1],LocalMemory[X][Event_object[X][Y][ E-1]])
										}


										  /*-----------------check the opacity condition---------*/
							  else if   Event[X][Y][n]= res_read_OG and Event_value[X][Y][n]~=GlobalMemory[Event_object[X][Y][n]]
											then{
													Event[X][Y][E]:= res_abort and
													transaction_abort(X,Y,E)
													}
											else transaction_response_Read(X,Y,E,n-1)

}and


 /* some help functions regarding output
  *
*/


define out_operation(X,Y,operation,obj,value) = {

	 if operation=write then {
	 format("Process %t: Write(%s,%t,[P%t][T%t])\n",X,str_obj(obj),value,X,Y)
	 } else if operation=read then{
	 format("Process %t: Read(%s,[P%t][T%t])..\n",X,str_obj(obj),X,Y)
	 } else if operation=try_commit then{
	 format("Process %t: TryCommit([P%t][T%t])\n",X,X,Y)
	 } else if operation=try_abort then{
	 format("Process %t: TryAbort([P%t][T%t])\n",X,X,Y)
	 }else if operation=res_commit then{
	 format("Process %t: Commit([P%t][T%t])\n",X,X,Y)
	 } else if operation=res_abort then{
	 format("Process %t: Abort([P%t][T%t])\n",X,X,Y)
	 }else if operation=res_read_OL  then{
	 format("Process %t: res_read_LocMem(%s,%t,[P%t][T%t])\n",X,str_obj(obj),value,X,Y)
	 } else if  operation=res_read_OG then{
	 format("Process %t: res_read_GloMem(%s,%t,[P%t][T%t])\n",X,str_obj(obj),value,X,Y)
	 } else if operation=res_write_OK then{
	 format("Process %t: res_write_OK([P%t][T%t])\n",X,X,Y)
	 }


 } and

define str_obj(X) = {
  if X=a then "a"
  else if X=b then "b"
  else if X=c then "c"
	 else "in a not currently known state"
 } and



/*--------------------&&&&&&&&.........the main program.......&&&&&&-----------------------------------*/

 tcl("init",[nprocesses,ntransactions,nlocations]) and
 always tclbreak() and
 always(

		  tcl("tmr",[Timer]) and
		  (forall j<nlocations :tcl("globalMemory",[j,GlobalMemory[j]])) and
		  (forall i<nprocesses :
			 (forall j<nlocations :tcl("localMemory",[i,j,LocalMemory[i][j]]))) and
		  (forall i<nprocesses :
			(forall j<ntransactions :
				(forall k<nevents :tcl("event",[i,j,Event[i][j][k],Event_object[i][j][k],Event_value[i][j][k],k,Tr_begin_time[i][j]]))))


		) and

 Timer gets Timer+1 and


/*--------------------&&&&&&&&.........A sample run of the Transactions.......&&&&&&-----------------------------------*/


 {
/*--------------state 0: process 0: transaction_begin(deposit 100 to a)----------------------*/

	{

	(forall i<nprocesses   : if i~=0 then {process_idle(i)and itsParts_idle(i)}

												else{
													  forall j<ntransactions: if j~=0 then transactionEvent_idle(i,j)
																								  else {forall e<nevents: if e~=0 then Event_idle(i,j,e)
																																			 else transaction_begin(i,j,e,write,a,100)
																										  }
											 }



		)
  };

  {

	(forall i<nprocesses   : if i~=0 then {process_idle(i)and itsParts_idle(i)}

												else{
													  forall j<ntransactions: if j~=0 then transactionEvent_idle(i,j)
																								  else {forall e<nevents: if e~=1 then Event_idle(i,j,e)
																																			 else transaction_response(i,j,e)
																										  }
											 }



		)
  };

	 {

	(forall i<nprocesses   : if i~=0 then {process_idle(i)and itsParts_idle(i)}

												else{
													  forall j<ntransactions: if j~=0 then transactionEvent_idle(i,j)
																								  else {forall e<nevents: if e~=2 then Event_idle(i,j,e)
																																			 else transaction_end(i,j,e,try_commit)
																										  }
											 }



		)
  };
	 {

	(forall i<nprocesses   : if i~=0 then {process_idle(i)and itsParts_idle(i)}

												else{
													  forall j<ntransactions: if j~=0 then transactionEvent_idle(i,j)
																								  else {forall e<nevents: if e~=3 then Event_idle(i,j,e)
																																			 else transaction_response(i,j,e)
																										  }
											 }



		)
  };

/*--------------state 4: process 1: transaction_begin(transfare 100 to b)----------------------*/

	{

	(forall i<nprocesses   : if i~=1 then {process_idle(i)and itsParts_idle(i)}

												else{
													  forall j<ntransactions: if j~=0 then transactionEvent_idle(i,j)
																								  else {forall e<nevents: if e~=4 then Event_idle(i,j,e)
																																			 else transaction_begin(i,j,e,read,a,no_value)
																										  }
											 }



		)
  };

  {

	(forall i<nprocesses   : if i~=1 then {process_idle(i)and itsParts_idle(i)}

												else{
													  forall j<ntransactions: if j~=0 then transactionEvent_idle(i,j)
																								  else {forall e<nevents: if e~=5 then Event_idle(i,j,e)
																																			 else transaction_response(i,j,e)
																										  }
											 }



		)
  };

  {

	(forall i<nprocesses   : if i~=1 then {process_idle(i)and itsParts_idle(i)}

												else{
													  forall j<ntransactions: if j~=0 then transactionEvent_idle(i,j)
																								  else {forall e<nevents: if e~=6 then Event_idle(i,j,e)
																																			 else transaction_continue(i,j,e,write,a,0)
																										  }
											 }



		)
  };

  {

	(forall i<nprocesses   : if i~=1 then {process_idle(i)and itsParts_idle(i)}

												else{
													  forall j<ntransactions: if j~=0 then transactionEvent_idle(i,j)
																								  else {forall e<nevents: if e~=7 then Event_idle(i,j,e)
																																			 else transaction_response(i,j,e)
																										  }
											 }



		)
  };

  {

	(forall i<nprocesses   : if i~=1 then {process_idle(i)and itsParts_idle(i)}

												else{
													  forall j<ntransactions: if j~=0 then transactionEvent_idle(i,j)
																								  else {forall e<nevents: if e~=8 then Event_idle(i,j,e)
																																			 else transaction_continue(i,j,e,read,b,no_value)
																										  }
											 }



		)
  };

  {

	(forall i<nprocesses   : if i~=1 then {process_idle(i)and itsParts_idle(i)}

												else{
													  forall j<ntransactions: if j~=0 then transactionEvent_idle(i,j)
																								  else {forall e<nevents: if e~=9 then Event_idle(i,j,e)
																																			 else transaction_response(i,j,e)
																										  }
											 }



		)
  };

  {

	(forall i<nprocesses   : if i~=1 then {process_idle(i)and itsParts_idle(i)}

												else{
													  forall j<ntransactions: if j~=0 then transactionEvent_idle(i,j)
																								  else {forall e<nevents: if e~=10 then Event_idle(i,j,e)
																																			 else transaction_continue(i,j,e,write,b,100)
																										  }
											 }



		)
  };

  {

	(forall i<nprocesses   : if i~=1 then {process_idle(i)and itsParts_idle(i)}

												else{
													  forall j<ntransactions: if j~=0 then transactionEvent_idle(i,j)
																								  else {forall e<nevents: if e~=11 then Event_idle(i,j,e)
																																			 else transaction_response(i,j,e)
																										  }
											 }



		)
  };

   {

	(forall i<nprocesses   : if i~=1 then {process_idle(i)and itsParts_idle(i)}

												else{
													  forall j<ntransactions: if j~=0 then transactionEvent_idle(i,j)
																								  else {forall e<nevents: if e~=12 then Event_idle(i,j,e)
																																			 else transaction_end(i,j,e,try_commit)
																										  }
											 }



		)
  };
	 {

	(forall i<nprocesses   : if i~=1 then {process_idle(i)and itsParts_idle(i)}

												else{
													  forall j<ntransactions: if j~=0 then transactionEvent_idle(i,j)
																								  else {forall e<nevents: if e~=13 then Event_idle(i,j,e)
																																			 else transaction_response(i,j,e)
																										  }
											 }



		)
  }


}

 }

}.
set print_states=true.
