/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2017 Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: io.c 284 2017-10-08 08:00:31Z cau $
 */


#include "system.h"


/******************************\
*                              *
*  Auxiliary output routines.  *
*                              *
\******************************/

char *titoa(i)
int64_t i;
{
   static char buf[MAXINTSIZE];
   char *p = buf;
   snprintf(buf,MAXINTSIZE,"%" PRId64 "",i);
   return p;
 }

void output_int(n,f)
int64_t n;
int f;
{
    char    s[MAXINTSIZE],
           *p = s;
    snprintf(s,MAXINTSIZE,"%" PRId64 "",n);
    outputf_str (p, f);
}

void foutput_int(n,f)
int64_t n;
int f;
{
    char    s[MAXINTSIZE],
            *p=s;
    
    snprintf(s,MAXINTSIZE,"%" PRId64 "",n);
    foutputf_str (p, f);
}


void output_str(s)
char *s;
{
    int c;
    
    while (*s != '\0') {
        if (DebugLevel>2) {
            switch (c=*s++) {
                case '\n' :
                    output_char('\\');
                    output_char('n');
                    break;
                case '\r' :
                    output_char('\\');
                    output_char('r');
                    break;
                case '\t' :
                    output_char('\\');
                    output_char('t');
                    break;
                case '\b' :
                    output_char('\\');
                    output_char('b');
                    break;
                case '\f' :
                    output_char('\\');
                    output_char('f');
                    break;
                case '\\' :
                    output_char('\\');
                    output_char('\\');
                    break;
                case '"' :
                    output_char('\\');
                    output_char('"');
                    break;
                default:
                    output_char (c);
                    break;
            }
        }
        else
            output_char (*s++);
    }
    
}

void foutput_str(s)
char *s;
{
    int c;
    while (*s != '\0') {
        if (DebugLevel>2) {
            switch (c=*s++) {
                case '\n' :
                    foutput_char('\\');
                    foutput_char('n');
                    break;
                case '\r' :
                    foutput_char('\\');
                    foutput_char('r');
                    break;
                case '\t' :
                    foutput_char('\\');
                    foutput_char('t');
                    break;
                case '\b' :
                    foutput_char('\\');
                    foutput_char('b');
                    break;
                case '\f' :
                    foutput_char('\\');
                    foutput_char('f');
                    break;
                case '\\' :
                    foutput_char('\\');
                    foutput_char('\\');
                    break;
                case '"' :
                    foutput_char('\\');
                    foutput_char('"');
                    break;    
                default:
                    foutput_char (c);
                    break;
            }
        }
        else
            foutput_char (*s++);
    }
}

void output_fl(n)
double n;
{
    char    s[MAXFLSIZE];
    char *s2,*fmt,*p;
    
    s2 = strsave(titoa(Precision));
    p = strapnd(s2,"e");
    fmt = strapnd("%.",p);
    snprintf(s,MAXFLSIZE-1,fmt,n);
    free_mem(fmt);
    free_mem(p);
    free_mem(s2);
    output_str(s);
}

void foutput_fl(n)
double n;
{
    char    s[MAXFLSIZE];
    char *s2,*fmt,*p;
    
    s2 = strsave(titoa(Precision));
    p = strapnd(s2,"e");
    fmt = strapnd("%.",p);
    snprintf(s,MAXFLSIZE-1,fmt,n);
    free_mem(fmt);
    free_mem(p);
    free_mem(s2);
    foutput_str(s);
}

void outputf_str(s,f)
char *s;
int f;
{
    int     len = strlen (s);
    spaces (f - len);
    output_str (s);
    spaces (-f - len);
}

void foutputf_str(s,f)
char *s;
int f;
{
    int     len = strlen (s);
    fspaces (f - len);
    foutput_str (s);
    fspaces (-f - len);
}

void outputf_fl(n,f)
double n;
int f;
{
    char    s[MAXFLSIZE];
    char *s1,*s2,*fmt,*p,*p1,*p2;

    s1 = strsave(titoa(f));
    s2 = strsave(titoa(Precision));
    p = strapnd(s2,"e");
    p1 = strapnd(".",p);
    p2 = strapnd(s1,p1);
    fmt = strapnd("%",p2);
    snprintf(s,MAXFLSIZE-1,fmt,n);
    free_mem(fmt);
    free_mem(p2);
    free_mem(p1);
    free_mem(p);
    free_mem(s2);
    free_mem(s1);
    outputf_str(s,f);
}

void foutputf_fl(n,f)
double n;
int f;
{
    char    s[MAXFLSIZE];
    char *s1,*s2,*fmt,*p,*p1,*p2;

    s1 = strsave(titoa(f));
    s2 = strsave(titoa(Precision));
    p = strapnd(s2,"e");
    p1 = strapnd(".",p);
    p2 = strapnd(s1,p1);
    fmt = strapnd("%",p2);
    snprintf(s,MAXFLSIZE-1,fmt,n);
    free_mem(fmt);
    free_mem(p2);
    free_mem(p1);
    free_mem(p);
    free_mem(s2);
    free_mem(s1);
    foutputf_str(s,f);
}

void foutputf_fl_f(n,f)
double n;
int f;
{
    char    s[MAXFLSIZE];
    char *s1,*s2,*fmt,*p,*p1,*p2;
    
    s1 = strsave(titoa(f));
    s2 = strsave(titoa(Precision));
    p = strapnd(s2,"f");
    p1 = strapnd(".",p);
    p2 = strapnd(s1,p1);
    fmt = strapnd("%",p2);
    snprintf(s,MAXFLSIZE-1,fmt,n);
    free_mem(fmt);
    free_mem(p2);
    free_mem(p1);
    free_mem(p);
    free_mem(s2);
    free_mem(s1);
    foutputf_str(s,f);
}

void output_lambda(e)
NODEPTR e;
{
    output_expn(e,-1);
}

void foutput_lambda(e)
NODEPTR e;
{
    foutput_expn(e,-1);
}

/* Find print-name of type. */
char *print_name(e)
NODEPTR e;
{
    unsigned t = node_type(e) == 0 ? token_type(END) : token_type(node_type(e));
    static char *undef = "undefined";
    return ((t <= token_type(MAXTOKEN)) ? Keyword[t] : undef);
}

void message(text)
char *text;
{
    printf ("***Tempura: %s.\n", text);
    fflush(stdout);
}

void spaces(n)
int n;
{
    while (n-- > 0)
        output_char (' ');
}

void fspaces(n)
int n;
{
    while (n-- > 0)
        foutput_char (' ');
}

void output_env(env)
NODEPTR env;
{
    NODEPTR id = node_env_idlist(env);
    while (id != NULL) {
        output_str (node_nm_name(id));
        if ((id = node_next(id)) != NULL)
            output_str (", ");
    }
}

void foutput_env(env)
NODEPTR env;
{
    NODEPTR id = node_env_idlist(env);
    while (id != NULL) {
        foutput_str (node_nm_name(id));
        if ((id = node_next(id)) != NULL)
            foutput_str (", ");
    }
}

void output_expn(e,level)
NODEPTR e;
int level;
{

    if (e == NULL) {
/*        output_str ("()");*/
        output_str("?");
        return;
    }

    if (level == 0) {
        output_str ("...");
        return;
    }

    switch (node_type(e)) {
        
        case NAME:
            output_char ('(');
            output_str (Keyword[token_type(node_type(e))]);
            space;
            output_str (node_nm_name(e));
            output_char (')');
            break;
        case STATE:
        case STATIC:
            output_str (node_nm_name(e));
            break;
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case LIST:
        case LOC:
            output_val (e);
            break;
        case BND:
            output_expn (node_bnd_env(e), level); 
            break;
        case ENV:
            output_env (e); 
            break;
        case MEMBLOCK:
            output_str (Keyword[token_type(node_type(e))]);
            output_int ((signed)node_blk_size(e), -1);
            break;
        case SKIP:
        case EMPTY:
        case INFINITE:
        case RANLEN:    
        case MORE:    
            output_str (Keyword[token_type(node_type(e))]);
            break;
        case LT:
        case LE:
        case GE:
        case GT:
        case EQ:
        case NEQ:
        case PLUS:
        case MINUS:
        case MULT:
        case DIV:
        case POWER:
        case AND:
        case OR:
        case IMPLIES:
        case MOD:
            output_char ('(');
            output_expn (node_lnk_1(e), level);
	    space;
            output_str (Keyword[token_type(node_type(e))]);
	    space;
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;
        case ATANTWO:
            output_str (Keyword[token_type(node_type(e))]);
            output_char ('(');
            output_expn (node_lnk_1(e), level);
            output_char (',');
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;    
        case SEMI:
        case WHERE:
        case PROJ:
        case GETS:
        case IS:
        case ASSN:
        case UASSN:
        case SKIPUASSN:    
        case DEFAULTS:
            output_expn (node_lnk_1(e), level);
	    space;
            output_str (Keyword[token_type(node_type(e))]);
	    space;
            output_expn (node_lnk_2(e), level - 1);
            break;
        case iUNDEFINED:
            output_str (Keyword[token_type(node_type(e))]);
            output_char ('(');
            output_expn (node_lnk_1(e), level);
            output_char (')');
            break;
        case NOT:
        case UPLUS:
        case UMINUS:
        case REFER:
        case DEREF:    
        case PROCESS:
        case iPROCESS:
        case PREFIX:
        case iPREFIX:
        case ALWAYS:
        case SOMETIME:
        case HALT:
        case INIT:
        case FIN:
        case CHOPSTAR:    
        case KEEP:
        case iLEN:    
        case STB:
        case MEM:
        case INPUT:
        case WINPUT:
        case OUTPUT:
        case iREMEMBER:
        case RUN:
        case TYPE:
        case ASCII:
        case STRUCT:
        case LOAD:
        case SYSTEM:
            output_str (Keyword[token_type(node_type(e))]);
	    space;
            output_expn (node_lnk_1(e), level);
            break;
        case NEXT:
        case WNEXT:
        case LEN:    
        case PREV:    
        case CEIL:
        case FLOOR:
        case SQRT:
        case ITOF:
        case EXPO:
        case LOG:
        case LOGTEN:
        case SIN:
        case COS:
        case TAN:
        case ASIN:
        case ACOS:
        case ATAN:
        case SINH:
        case COSH:
        case TANH:
        case FABS:
        case FORMAT:
            output_str (Keyword[token_type(node_type(e))]);
	    output_char ('(');;
            output_expn (node_lnk_1(e), level);
	    output_char (')');;
            break;
        case IF:
        case iIF:    
            output_str (Keyword[token_type(node_type(e))]);
            space;
            output_expn (node_lnk_1(e), level);
            output_str (" then (");
            output_expn (node_lnk_2(e), level - 1);
            if (node_lnk_3(e) != NULL) {
                output_str (") else (");
                output_expn (node_lnk_3(e), level - 1);
            }
            output_char (')');
            break;
        case CALL:
        case iCALL:
            output_expn (node_lnk_1(e), level);
            output_char ('(');
            output_expn (node_lnk_2(e), level);
            output_char (')');
            break;
        case LAMBDA:
            output_str (Keyword[token_type(node_type(e))]);
            output_str (" (");
            output_expn (node_lnk_1(e), level);
            output_str (") : (");
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;
        case EXISTS:
        case FORALL:
            output_str (Keyword[token_type(node_type(e))]);
            space;
            output_expn (node_lnk_1(e), level);
            if (node_lnk_2(e) != NULL) {
                output_char ('<');
                output_expn (node_lnk_2(e), level);
            }
            output_str (": (");
            output_expn (node_lnk_3(e), level - 1);
            output_char (')');
            break;
        case WHILE:
            output_str (Keyword[token_type(node_type(e))]);
            output_char (' ');
            output_expn (node_lnk_1(e), level);
            output_str (" do (");
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;
        case FOR:
            output_str (Keyword[token_type(node_type(e))]);
            space;
            output_expn (node_lnk_1(e), level);
            output_char ('<');
            output_expn (node_lnk_2(e), level);
            output_str (" do (");
            output_expn (node_lnk_3(e), level - 1);
            output_char (')');
            break;
        case FORIN:
            output_str (Keyword[token_type(node_type(e))]);
            space;
            output_expn (node_lnk_1(e), level);
            output_str (" in ");
            output_expn (node_lnk_2(e), level);
            output_str (" do (");
            output_expn (node_lnk_3(e), level - 1);
            output_char (')');
            break;
        case FORTIMES:
            output_str ("for ");
            output_expn (node_lnk_1(e), level);
            output_str (" times do (");
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;
        case DEFINE:
        case SET:
            output_str ("(");
            output_str (Keyword[token_type(node_type(e))]);
            space;
            output_expn (node_lnk_1(e), level);
            output_char ('=');
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;
        case COMMA:
	    output_expn (node_lnk_1(e), level);
            if (node_lnk_2(e) != NULL) {
                output_char (',');
		output_expn (node_lnk_2(e), level);
            }
            break;
        case iEXISTS:
            output_str (Keyword[token_type(node_type(e))]);
            space;
            output_expn (node_lnk_1(e), level);
            output_str (" : (");
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;
        case iAND:
        case iOR:
        case iSEMI:
        case iPROJ:
            output_expn (node_lnk_2(e), level);
            if (node_lnk_3(e) != NULL) {
                space;
                output_str (Keyword[token_type(node_type(e))]);
                space;
                output_expn (node_lnk_3(e), level - 1);
            }
            break;
        case MAKELIST:
            output_char ('[');
            output_expn (node_lnk_1(e), level - 1);
            output_char (']');
            break;
        case SUB:
            output_expn (node_lnk_1(e), level);
            output_char ('[');
            output_expn (node_lnk_2(e), level - 1);
            output_char (']');
            break;
        case SIZE:
            output_char ('|');
            output_expn (node_lnk_1(e), level);
            output_char ('|');
            break;
        case SUBLIST:
            output_expn (node_lnk_1(e), level);
            output_char ('[');
            output_expn (node_lnk_2(e), level - 1);
            output_str ("..");
            output_expn (node_lnk_3(e), level - 1);
            output_char (']');
            break;
        case FIXLIST:
        case VARLIST:
        case NPREV:
        case iSYSTEM:
            output_str (Keyword[token_type(node_type(e))]);
            output_char ('(');
            output_expn (node_lnk_1(e), level - 1);
            output_char (',');
            output_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;    
        default:
            output_str (print_name (e));
            break;
    }
}

void foutput_expn(e,level)
NODEPTR e;
int level;
{

    if (e == NULL) {
/*        output_str ("()");*/
        foutput_str ("?");
        return;
    }

    if (level == 0) {
        foutput_str ("...");
        return;
    }

    switch (node_type(e)) {

        case STATE:
        case STATIC:
            foutput_str (node_nm_name(e));
            break;
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case LIST:
        case LOC:
            foutput_val (e);
            break;
        case BND:
            foutput_expn (node_bnd_env(e), level); 
            break;
        case ENV:
            foutput_env (e); 
            break;
        case MEMBLOCK:
            foutput_str (Keyword[token_type(node_type(e))]);
            foutput_int ((signed)node_blk_size(e), -1);
            break;
        case SKIP:
        case EMPTY:
        case INFINITE:
        case RANLEN:
        case MORE:    
            foutput_str (Keyword[token_type(node_type(e))]);
            break;
        case LT:
        case LE:
        case GE:
        case GT:
        case EQ:
        case NEQ:
        case PLUS:
        case MINUS:
        case MULT:
        case DIV:
        case POWER:
        case AND:
        case OR:
        case IMPLIES:
        case MOD:
            foutput_char ('(');
            foutput_expn (node_lnk_1(e), level);
	    fspace;
            foutput_str (Keyword[token_type(node_type(e))]);
	    fspace;
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (')');
            break;
        case ATANTWO:
            foutput_str (Keyword[token_type(node_type(e))]);
            foutput_char ('(');
            foutput_expn (node_lnk_1(e), level);
            foutput_char (',');
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (')');
            break;
        case SEMI:
        case WHERE:
        case PROJ:
        case GETS:
        case IS:
        case ASSN:
        case UASSN:
        case SKIPUASSN:    
        case DEFAULTS:
            foutput_expn (node_lnk_1(e), level);
	    fspace;
            foutput_str (Keyword[token_type(node_type(e))]);
	    fspace;
            foutput_expn (node_lnk_2(e), level - 1);
            break;
        case iUNDEFINED:
            foutput_str (Keyword[token_type(node_type(e))]);
            foutput_char ('(');
            foutput_expn (node_lnk_1(e), level);
            foutput_char (')');
            break;    
        case NOT:
        case UPLUS:
        case UMINUS:
        case REFER:
        case DEREF:    
        case PROCESS:
        case iPROCESS:
        case PREFIX:
        case iPREFIX:
        case ALWAYS:
        case SOMETIME:
        case HALT:
        case INIT:
        case FIN:
        case CHOPSTAR:    
        case KEEP:
        case iLEN:
        case STB:
        case MEM:
        case INPUT:
        case WINPUT:
        case OUTPUT:
        case iREMEMBER:
        case RUN:
        case TYPE:
        case ASCII:
        case STRUCT:
        case LOAD:
        case SYSTEM:
            foutput_str (Keyword[token_type(node_type(e))]);
	    fspace;
            foutput_expn (node_lnk_1(e), level);
            break;
        case NEXT:
        case WNEXT:
        case LEN:    
        case PREV:    
        case CEIL:
        case FLOOR:
        case SQRT:
        case ITOF:
        case EXPO:
        case LOG:
        case LOGTEN:
        case SIN:
        case COS:
        case TAN:
        case ASIN:
        case ACOS:
        case ATAN:
        case SINH:
        case COSH:
        case TANH:
        case FABS:
        case FORMAT:
            foutput_str (Keyword[token_type(node_type(e))]);
	    foutput_char ('(');;
            foutput_expn (node_lnk_1(e), level);
	    foutput_char (')');;
            break;
        case IF:
        case iIF:    
            foutput_str (Keyword[token_type(node_type(e))]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" then (");
            foutput_expn (node_lnk_2(e), level - 1);
            if (node_lnk_3(e) != NULL) {
                foutput_str (") else (");
                foutput_expn (node_lnk_3(e), level - 1);
            }
            foutput_char (')');
            break;
        case CALL:
        case iCALL:
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('(');
            foutput_expn (node_lnk_2(e), level);
            foutput_char (')');
            break;
        case LAMBDA:
            foutput_str (Keyword[token_type(node_type(e))]);
            foutput_str (" (");
            foutput_expn (node_lnk_1(e), level);
            foutput_str (") : (");
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (')');
            break;
        case EXISTS:
        case FORALL:
            foutput_str (Keyword[token_type(node_type(e))]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            if (node_lnk_2(e) != NULL) {
                foutput_char ('<');
                foutput_expn (node_lnk_2(e), level);
            }
            foutput_str (": (");
            foutput_expn (node_lnk_3(e), level - 1);
            foutput_char (')');
            break;
        case WHILE:
            foutput_str (Keyword[token_type(node_type(e))]);
            foutput_char (' ');
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" do (");
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (')');
            break;
        case FOR:
            foutput_str (Keyword[token_type(node_type(e))]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('<');
            foutput_expn (node_lnk_2(e), level);
            foutput_str (" do (");
            foutput_expn (node_lnk_3(e), level - 1);
            foutput_char (')');
            break;
        case FORIN:
            foutput_str (Keyword[token_type(node_type(e))]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" in ");
            foutput_expn (node_lnk_2(e), level);
            foutput_str (" do (");
            foutput_expn (node_lnk_3(e), level - 1);
            foutput_char (')');
            break;
        case FORTIMES:
            foutput_str ("for ");
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" times do (");
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (')');
            break;
        case DEFINE:
        case SET:
            output_char ('(');
            foutput_str (Keyword[token_type(node_type(e))]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('=');
            foutput_expn (node_lnk_2(e), level - 1);
            output_char (')');
            break;
        case COMMA:
	    foutput_expn (node_lnk_1(e), level);
            if (node_lnk_2(e) != NULL) {
                foutput_char (',');
		foutput_expn (node_lnk_2(e), level);
            }
            break;
        case iEXISTS:
            foutput_str (Keyword[token_type(node_type(e))]);
            fspace;
            foutput_expn (node_lnk_1(e), level);
            foutput_str (" : (");
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (')');
            break;
        case iAND:
        case iOR:
        case iSEMI:
        case iPROJ:
            foutput_expn (node_lnk_2(e), level);
            if (node_lnk_3(e) != NULL) {
                fspace;
                foutput_str (Keyword[token_type(node_type(e))]);
                fspace;
                foutput_expn (node_lnk_3(e), level - 1);
            }
            break;
        case MAKELIST:
            foutput_char ('[');
            foutput_expn (node_lnk_1(e), level - 1);
            foutput_char (']');
            break;
        case SUB:
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('[');
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (']');
            break;
        case SIZE:
            foutput_char ('|');
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('|');
            break;
        case SUBLIST:
            foutput_expn (node_lnk_1(e), level);
            foutput_char ('[');
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_str ("..");
            foutput_expn (node_lnk_3(e), level - 1);
            foutput_char (']');
            break;
        case FIXLIST:
        case VARLIST:
        case NPREV:
        case iSYSTEM:
            foutput_str (Keyword[token_type(node_type(e))]);
            foutput_char ('(');
            foutput_expn (node_lnk_1(e), level - 1);
            foutput_char (',');
            foutput_expn (node_lnk_2(e), level - 1);
            foutput_char (')');
            break;    
        default:
            foutput_str (print_name (e));
            break;
    }
}

void foutput_char (c)
int c;
{
    char   name[MAXFILENAMESIZE];
    FILE * fp = NULL,
         *openfile ();
    if (strcmp (node_strval(OutFile), "stdout") == 0) {
        output_char (c);
    }
    else {
        if ((fp = openfile (name, node_strval(Path), node_strval(OutFile), FXOUT, "a")) == NULL) {
            warn1 ("cannot open file \"%s\"",  node_strval(OutFile));
        warn0 ("will display to stdout");
        output_char (c);
        }
        else {
            putc(c,fp);
           CharsOutput++;
           fclose (fp);
        }
    }
}
