/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2017 Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: system.h 284 2017-10-08 08:00:31Z cau $
 */
#include "config.h"
#include <stdio.h>
#include <ctype.h>
#include <setjmp.h>
#include <signal.h>
#include <limits.h>
#include <math.h>
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#include <sys/types.h>
#include <unistd.h>
#include <stdint.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <time.h>
#include <sys/time.h>
#include <inttypes.h>
#include "tempura.h"
#include "parser.h"
#include "globals.h"

/*#define YYERROR_VERBOSE 1*/
#define YYDEBUG 1
#define YYTOKEN_TABLE 1

#ifdef _AIX
#   define WAIT_STATUS_TYPE pid_t
#else
#ifndef NO_UNION_WAIT
#if defined(__APPLE__) && defined(__MACH__)
#   define WAIT_STATUS_TYPE int
#else
#   define WAIT_STATUS_TYPE union wait
#endif
#else
#   define WAIT_STATUS_TYPE int
#endif
#endif

/*
#if defined (__MINGW) || defined(__MINGW32__)
double drand48(void);
double erand48(unsigned short xsubi[3]);
long int lrand48(void);
long int nrand48(unsigned short xsubi[3]);
long int mrand48(void);
long int jrand48(unsigned short xsubi[3]);
void srand48(long int seedval);
unsigned short *seed48(unsigned short seed16v[3]);
void lcong48(unsigned short param[7]);
#endif
*/

#define free_mem(A) if (A != NULL) free (A)

#define resetin(F) if (feof(F)) {(F)->_cnt=0;(F)->_flag^=_IOEOF;}
/*#define resetin(F)*/

/*#define MAXSYSVAR 9*/
#define HASHTABLESIZE 400


/* For the VAX, SUN, etc: */
#define MAXHISTORY 25
#define MAXSYMBOLTABLESIZE 1000
#define MAXHEAPSIZE 20000
#define GLOBALMEMSIZE 5000
#define MAXENVSTACKSIZE 400
#define MAXREDUCESTACKSIZE 1000
#define MAXTOKENSIZE 256
#define MAXLINESIZE 1024
#define MAXFILENAMESIZE 256
#define MAXINTSIZE 21
#define MAXFLSIZE 32
#define FILENAMESEP '/'
#define FX ".t"
#define FXOUT ""
#define FXIN ""
#define HELPSUMMARY "summary"

/* For the PC:
#define MAXSYMBOLTABLESIZE 300
#define MAXHEAPSIZE 3200
#define GLOBALMEMSIZE 200
#define MAXENVSTACKSIZE 100
#define MAXREDUCESTACKSIZE 100
#define MAXTOKENSIZE 257
#define MAXFILENAMESIZE 40
#define MAXINTSIZE 20
#define FILENAMESEP '\\'
#define FX ".t"
#define HELPSUMMARY "summary"
#define CNTRL_C 0x23
#define INTERRUPTSTACK 128
*/
